library(argparse)
library(readr)
suppressPackageStartupMessages(library(plotly))
# library(moments) #skewness & kurtosis

# TODO get ride of "TA_counts" col, use index instead
# TODO nous n'acceptons que les fichiers wigs avec un espace comem sep

ignore_this <-
  function(w) {
    if (any(grepl("'hoveron'", w))) {
      invokeRestart("muffleWarning")
    }
  }

parse_wig <- function(path, verbose = T) {
  # Parse the wig file and return a dataframe
  # ordered by ascending TA sites positions on genome
  if (!file.exists(path)) {
    stop(paste("Input wig file does not exist:", path))
  }

  # determine the number of line to skip
  lineToSkip <- 0
  conn <- file(path, open = "r")
  lines <- readLines(conn)
  for (i in 1:length(lines)) {
    line <- trimws(lines[i])
    firstCharacter <- substr(line, 1, 1)
    if ((firstCharacter %in% 1:9)) {
      # for r '1' %in% 1:9 is equivalent to 1 %in% 1:9
      break
    } else {
      lineToSkip <- lineToSkip + 1
    }
  }
  close(conn)
  if (lineToSkip != 0) {
    if (verbose) {
      print(paste(
        "warning: ignoring",
        lineToSkip,
        "header line(s) from wiggle:",
        path
      ))
    }
  }
  df <- read_delim(
    path,
    delim = " ",
    col_names = c("TA_pos", "TA_counts"),
    skip = lineToSkip,
    show_col_types = F,
    progress = F
  )
  # sort by ascending position
  # df = df[order(df$TA_pos),]
  df <- arrange(df, TA_pos)
  return(df)
}



# --

# LOESS --
perform_loess_prediction <- function(df) {
  df.loess <- loess(
    formula = TA_counts ~ TA_pos,
    data = df,
    span = 1,
    control = loess.control(
      statistics = c("approximate"),
      trace.hat = c("approximate")
    )
  )
  pred <- predict(df.loess, data.frame(TA_pos = df$TA_pos))
  return(pred)
}

perform_loess_normalization <- function(df, pred = NULL) {
  if (is.null(pred)) {
    pred <- perform_loess_prediction(df)
  }
  ratio <- pred / median(pred)
  df <- df %>% mutate(TA_counts = TA_counts / ratio)
  # df[, "TA_counts"] = df[, "TA_counts"] / ratio
  return(df)
}

# --
write_wig <- function(df, path, fields_separator = " ") {
  options(scipen=999) # avoid scientific notation as "transit normalize" don't like it
  write_delim(df, file=path,
    delim = fields_separator,
    quote = "none", col_names = FALSE,
    append = FALSE
  )
  options(scipen=0) # reverse
}


qc_pearson <-
  function(wig1, wig2, name1, name2, outdir, verbose = T) {
    if (!file.exists(wig1)) {
      stop("Input file for condition1 does not exist")
    }
    if (!file.exists(wig2)) {
      stop("Input file for condition2 does not exist")
    }

    data1 <- parse_wig(wig1, verbose = verbose)
    data2 <- parse_wig(wig2, verbose = verbose)

    if (!all(data1$TA_pos == data2$TA_pos)) {
      stop("TA positions are not the same between the two wiggle files.")
    }

    comp <- data.frame(c1 = data1$TA_counts, c2 = data2$TA_counts)

    pcor <-
      cor(data1$TA_counts, data2$TA_counts, method = "pearson")

    # print(paste("Pearson correlation:", pcor))

    p <- ggplot(comp, aes(c1, c2)) +
      geom_point(col = "steelblue") +
      labs(
        title = paste0(
          "Comparaison of the insertions count in two wiggle files (r=",
          signif(pcor, 2),
          ")"
        ),
        x = name1,
        y = name2
      )

    fig <- plotly::toWebGL(p) # manage huge dataset
    # fig <- ggplotly(p)

    # write plot
    if (!dir.exists(outdir)) {
      dir.create(outdir, recursive = TRUE)
    }

    # save Widget to no accept relative path
    hist_path <- file.path(normalizePath(outdir), "compare.html")
    lib_dir <- file.path(normalizePath(outdir), "lib")

    withCallingHandlers(
      htmlwidgets::saveWidget(fig,
        hist_path,
        selfcontained = F,
        libdir = lib_dir
      ),
      warning = ignore_this
    )

    return(list(pearson_correlation = pcor, outfile = hist_path))
  }






volcano_plot_from_transit_resampling_result_file <- function(input,
                                                             outdir,
                                                             qvalue_threshold,
                                                             log2fc_threshold) {
  qval_colname <- "Adj. p-value"
  logfc_colname <- "log2FC"
  locus_tag_colname <- "#Orf"
  filename <- basename(dirname(input))
  df <- parse_transit_resampling_file(input)
  return(
    volcano_plot(
      df,
      filename,
      outdir,
      qvalue_threshold,
      log2fc_threshold,
      qval_colname,
      logfc_colname,
      locus_tag_colname
    )
  )
}

# TODO: à reverifier
volcano_plot <-
  function(res,
           filename,
           outdir,
           qvalue_threshold,
           log2fc_threshold,
           qval_colname,
           logfc_colname,
           locus_tag_colname) {
    # res %>% colnames(.) %>% print()
    # rename columns for convenience (new name on left)
    # more on .env notation : https://tinyheero.github.io/2020/03/01/use-data-env-pronouns-tidyverse.html
    res <-
      res %>%
      rename(
        "log2FoldChange" = .env[["logfc_colname"]],
        "padj" = .env[["qval_colname"]],
        "locus_tag" = .env[["locus_tag_colname"]]
      ) %>%
      select(locus_tag, log2FoldChange, padj)

    res <- na.omit(res)
    res <- res %>% mutate(padj = ifelse(padj == 0, .Machine$double.xmin, padj))

    padj.threshold <-
      qvalue_threshold # Threshold on the adjusted p-value
    log2FoldChange.threshold <-
      log2fc_threshold # Threshold on the log2FoldChange
    cols <-
      c(
        "Down_regulated" = "cornflowerblue",
        "Not_significant" = "grey",
        "Up_regulated" = "firebrick"
      )

    # Generate an interactive volcano plot with plotly
    res$sig <-
      ifelse((
        res$log2FoldChange <= -log2FoldChange.threshold &
          res$padj <= padj.threshold
      ),
      "Down_regulated",
      ifelse((
        res$log2FoldChange >= log2FoldChange.threshold &
          res$padj <= padj.threshold
      ),
      "Up_regulated",
      "Not_significant"
      )
      )
    # write_csv(res, paste0("/home/rstudio/", filename, '.csv'))
    vline <- function(x = 0, color = "lightgrey") {
      list(
        type = "line",
        y0 = 0,
        y1 = 1,
        yref = "paper",
        x0 = x,
        x1 = x,
        line = list(color = color, dash = "dash")
      )
    }
    hline <- function(y = 0, color = "lightgrey") {
      list(
        type = "line",
        x0 = 0,
        x1 = 1,
        xref = "paper",
        y0 = y,
        y1 = y,
        line = list(color = color, dash = "dash")
      )
    }

    p <- plot_ly(
      data = res,
      y = ~ -log10(padj),
      x = ~log2FoldChange,
      color = ~sig,
      colors = cols,
      text = ~locus_tag
    ) %>%
      add_markers() %>%
      layout(
        title = paste(
          "Volcano plot on ",
          filename,
          "\n",
          "Significant genes: qvalue ≤ ",
          qvalue_threshold,
          "abs(log2fc) ≥ ",
          log2fc_threshold
        ),
        shapes = list(
          vline(-log2FoldChange.threshold),
          vline(log2FoldChange.threshold),
          hline(-log10(padj.threshold))
        )
      )

    fig <- ggplotly(p)

    # write volcano plot
    if (!dir.exists(outdir)) {
      dir.create(outdir, recursive = TRUE)
    }

    # save Widget to no accept relative path
    hist_path <- file.path(normalizePath(outdir), "volcano.html")
    lib_dir <- file.path(normalizePath(outdir), "lib")


    htmlwidgets::saveWidget(fig,
      hist_path,
      selfcontained = F,
      libdir = lib_dir
    )

    return(hist_path)
  }

parse_transit_resampling_file <- function(path) {
  df <- read_delim(
    path,
    delim = "\t",
    escape_double = FALSE,
    trim_ws = TRUE,
    skip = 7,
    show_col_types = FALSE
  )
  return(df)
}


pval_hist <-
  function(df,
           filename,
           outdir,
           nb_intervals,
           pval_colname) {
    # TODO what if the colname does not exist ?
    df <- df[!is.na(df[[pval_colname]]), ] # drop row with pval == NA

    # number of intervals to use
    if (is.null(nb_intervals)) {
      n <- nclass.Sturges(df[[pval_colname]])
    } else {
      n <- nb_intervals
      # print(paste("Number of desired intervals:", n))
    }

    # buid the histogram
    breaks <- pretty(range(df[[pval_colname]]),
      n = n,
      min.n = 1
    )

    p <- ggplot(df, aes(x = get(pval_colname))) +
      geom_histogram(breaks = breaks, fill = "#75AADB") +
      labs(
        title = paste(
          "Histogram of raw p-values on ",
          filename,
          "(Number of desired intervals:",
          n,
          ")"
        ),
        x = "p-values",
        y = "Frequency"
      )
    fig <- ggplotly(p)

    # write histogram
    if (!dir.exists(outdir)) {
      dir.create(outdir, recursive = TRUE)
    }

    # save Widget to no accept relative path
    hist_path <- file.path(normalizePath(outdir), "histogram.html")
    lib_dir <- file.path(normalizePath(outdir), "lib")

    htmlwidgets::saveWidget(fig,
      hist_path,
      selfcontained = F,
      libdir = lib_dir
    )
    return(hist_path)
  }

pval_hist_from_transit_resampling_result_file <-
  function(input, outdir, nb_intervals) {
    pval_colname <- "p-value"
    filename <- basename(dirname(input))
    df <- parse_transit_resampling_file(input)
    return(pval_hist(df, filename, outdir, nb_intervals, pval_colname))
  }
