library(shiny)
library(tidyverse)
library(plotly)
library(shinydashboard)
library(DT)
library(glue)
library(shinyvalidate)
library(shinybusy)
library(shinyFiles)

PLAYGROUND <- file.path(fs::path_home(), "playground")

# -----------------------------------------------------------------------------#
# HELPER FUNCTIONS
# -----------------------------------------------------------------------------#

discover_databases <- function(playground) {
  # ajout d'une chaine vide car le selectize input selectionne un element (sans qu'on lui demande).
  # permet de faire la différence entre cette selection auto et qd l'utilisteur fait l'action.
  c("",as.vector(list.files(path = playground, pattern = "\\.db$", full.names=TRUE, recursive = T)))
}

hmm_prediction_color <- styleEqual(
  c("ES", "NE", "GA", "GD", "N/A"),
  c("#FF934F", "#848FA2", "#CC2D35", "#058ED9", "#2D3142")
)
resampling_prediction_color <- styleEqual(
  c("unsignificant", "up", "down"),
  c("#848FA2", "#CC2D35", "#058ED9")
)

is_significant <- function(qval, logfc, qvalue_threshold, log2fc_threshold) {
  ifelse(
    (
      logfc <= -log2fc_threshold & qval <= qvalue_threshold
    ),
    "down",
    ifelse((
      logfc >= log2fc_threshold & qval <= qvalue_threshold
    ),
    "up",
    "unsignificant"
    )
  )
}

# ratio nb sites avec insertions / nb sites total
# sum(c(T,T,F,T)) --> 3
saturation_fun <- function(count) {
  sum(count > 0) / sum(count >= 0)
}

# visualisation des insertions pour une feature donnée
plot_feature_insertions <- function(table, ymax, start_on_ref_crit, end_on_ref_crit, how_to_handle_rep, show_legend=TRUE, make_it_relative=FALSE, strain=NULL) {
  # Make a plot in two parts. The upper part contains a bar plot of the insertions count for all conditions
  # The bottom contains the positions of TA site on the reference

  x0 <- start_on_ref_crit
  x1 <- end_on_ref_crit
  table <- table %>% mutate(absolute_position = position)
  if (make_it_relative) {
    table <- table %>% mutate(position = position - start_on_ref_crit)
    x0 <- 0
    x1 <- end_on_ref_crit - start_on_ref_crit
  }

  up <- plot_ly(data = table, showlegend = show_legend, legendgroup = ~condition) %>%
    add_bars(
      x = ~position,
      y = ~ get(how_to_handle_rep), # get pour récupérer le nom de la colone depuis une variable. Ce qui est affiché est la moyenne|sommme des comptages des replicats.
      color = ~condition,
      colors = "Set1",
      opacity = .5,
      width = .8,
      hoverinfo = 'text',
      text = ~ paste('Condition:', condition, '<br>',
                     'Absolute position:', absolute_position, '<br>',
                     '#insertions:', get(how_to_handle_rep), '<br>')
    ) %>%
    layout(barmode = "overlay", legend = list(title = list(text = "<b> Condition(s): </b>"), x = 1, y = 1))

  # Remove grid etc for the bottom part
  ax <- list(
    title = "",
    zeroline = TRUE,
    showline = FALSE,
    showticklabels = FALSE,
    showgrid = FALSE
  )

  ta <- table %>% mutate(count = -1)
  down <- plot_ly(data = ta, showlegend = F) %>%
    add_bars(
      x = ~position,
      y = ~count,
      color = I("black"),
      width = .8
    ) %>%
    layout(xaxis = ax, yaxis = ax, shapes = list(
      list( # ajout d'un rectangle pour matérialiser les positions de la feature courante
        type = "rect",
        fillcolor = "firebrick", line = list(color = "firebrick"), opacity = 0.2,
        x0 = x0, x1 = x1, xref = "x",
        y0 = 0, y1 = -1, yref = "y"
      )
    ))

  # Combine the up & down in one plot
  # fixedrange=T for yaxis to ensure that all plots keep the same y range (more comparable)
  fig <-
    subplot(
      up,
      down,
      nrows = 2,
      shareX = TRUE,
      heights = c(0.9, 0.1),
      margin = 0,
      titleX = FALSE,
      titleY = FALSE
    ) %>% layout(
      showlegend = T, # show even if there is only one condition
      bargap = 0,
      xaxis = list(title = "Positions (bp)"),
      yaxis = list(
        title = "#Insertions",
        fixedrange = TRUE,
        range = c(0, ymax)
      )
    )

  ## Put the name of the current strain un upper left corner of the plot
  if (!is.null(strain)) {
    fig <- fig %>% add_annotations(
      text = strain,
      x = 0.01,
      y = 1,
      yref = "paper",
      xref = "paper",
      xanchor = "middle",
      yanchor = "top",
      showarrow = FALSE,
      font = list(size = 10)
    )
  }
  fig

}
# -----------------------------------------------------------------------------#
# UI
# -----------------------------------------------------------------------------#

ui <- dashboardPage(
  header = dashboardHeader(title = "TnSeek"),
  sidebar = dashboardSidebar(
    sidebarMenu(
      id = "sidebarID",
      menuItem("Load data", tabName = "load_data"),
      menuItem("Strains overview", tabName = "strains"),
      menuItem("Discover locus tag",
        tabName = "discover_locus_tag"
      ),
      menuItem("Browse by locus tag",
        tabName = "locus_tag"
      ),
      menuItem("Browse by orthology group",
               tabName = "ortho_group"
      ),
      menuItem(
        "Global parameters",
        numericInput("log2fc_threshold", "Log2FC threshold:", value = 1),
        numericInput("qvalue_threshold", "Q-value threshold:", value = 0.05),
        selectInput(
          "how_to_handle_rep", "How to handle rep:",
          c(
            "Sum" = "sum",
            "Mean" = "mean"
          )
        )
      )
    )
  ),
  body = dashboardBody(
    add_busy_bar(color = "red", height = "8px"),
    tabItems(
      tabItem(
        tabName = "load_data",
        box(
          title = "Select database",
          width = 12,
          solidHeader = TRUE,
          status = "primary",
          selectInput("database", "Select your database:", choices = NULL, selected = NULL, multiple = FALSE)
          #inputFileUI("database"),
          #actionButton("db_connect", "Connect to selected database")
        ),
        box(
          title = "Current database",
          width = 12,
          solidHeader = TRUE,
          status = "primary",
          textOutput("current_database")
        )
        
      ),
      tabItem(
        tabName = "strains",
        column(
          width=8,
          box(
            title = "Strains overview",
            width = 12,
            solidHeader = TRUE,
            status = "primary",
            DT::dataTableOutput("strain_overview_table")
          ),
          box(
            title = "Distribution of TA number by feature",
            width = 12,
            solidHeader = TRUE,
            status = "primary",
            plotlyOutput("view")
          )
          
        ),
        column(
          width=4,
          infoBoxOutput("")
        ),
        column(
          width=4,
          box(
            title = "Distribution of feature number by orthologous group",
            width = 12,
            solidHeader = TRUE,
            status = "primary",
            p("Number of orthologous groups: ",textOutput("orthologous_groups_number", inline = T)),
            plotlyOutput("orthologous_groups_distribution")
          )

        )
      ),
      tabItem(
        tabName = "discover_locus_tag",
        fluidRow(
          box(
            title = "Advanced filters",
            width = 12,
            collapsible = TRUE,
            collapsed = TRUE,
            solidHeader = TRUE,
            status = "primary",
            checkboxInput("ignore_boring_rows", "Ignore rows with unsignificant values", value = FALSE),
            helpText("If a row has 'NE' (HMM) or 'unsignificant' (RESAMPLING) in all conditions, then, ignore the row.")
          )
        ),
        fluidRow(
          box(
            title = "HMM predictions",
            width = 12,
            collapsible = TRUE,
            solidHeader = TRUE,
            status = "primary",
            fluidRow(
              column(
                width = 12,
                DT::dataTableOutput("hmm_overall_prediction_table")
              )
            )
          )
        ),
        fluidRow(
          box(
            title = "Resampling predictions",
            width = 12,
            collapsible = TRUE,
            solidHeader = TRUE,
            status = "primary",
            fluidRow(
              column(
                width = 12,
                DT::dataTableOutput("resampling_overall_prediction_table")
              )
            )
          )
        )
      ),
      tabItem(
        tabName = "locus_tag",
        fluidRow(
          box(
            title = "Feature description",
            width = 12,
            solidHeader = TRUE,
            collapsible = TRUE,
            status = "primary",
            DT::dataTableOutput("feature_description_table")
          )
        ),
        fluidRow(
          box(
            title = "Selected locus tag",
            width = 12,
            solidHeader = TRUE,
            status = "primary",
            DT::dataTableOutput("selected_locus_tag_table")
          )
        ),
        fluidRow(
          box(
            title = "Insertions overview",
            width = 4,
            solidHeader = TRUE,
            status = "primary",
            DT::dataTableOutput("insertions_overview_table")
          ),
          box(
            title = "Predictions",
            width = 8,
            solidHeader = TRUE,
            status = "primary",
            fluidRow(
              column(
                width = 7,
                DT::dataTableOutput("hmm_predictions_table")
              ),
              column(
                width = 5,
                DT::dataTableOutput("resampling_predictions_table")
              )
            )
          )
        ),
        fluidRow(
          box(
            title = "Insertions visualization",
            width = 12,
            solidHeader = TRUE,
            status = "primary",
            column(
              2,
              numericInput("ymax_cutoff", "Ymax cutoff:", value = 150),
              numericInput("feature_context", "Context:", value = 0)
            ),
            column(10, plotlyOutput("insertions_plot")) # , height = '800'))
          )
        )
      ),
      tabItem(
        tabName = "ortho_group",
        fluidRow(
          box(
            title = "Insertions visualization",
            width = 12,
            collapsible = FALSE,
            collapsed = FALSE,
            solidHeader = TRUE,
            status = "primary",
            column(
              1,
              numericInput("ymax_cutoff_ortho", "Ymax cutoff:", value = 150)
            ),
            column(11, plotlyOutput("ortho_group_insertions_plot",height = '800'))
          )
        )
      )
    )
  )
)
# -----------------------------------------------------------------------------#
# SERVER
# -----------------------------------------------------------------------------#

server <- function(input, output, session) {
  #sqlitePath <- inputFileServer("database", filetypes=c('db'))
  # Connexion à la base de donnée
  updateSelectInput(session, 'database', choices = discover_databases(PLAYGROUND), selected = NULL)
  
  selected_database <- reactiveVal(NULL)
  con <- reactiveVal(NULL)

  observeEvent(input$database, {
    selected_db_rds <- "/tmp/selected_db.rds"
    # quand l'utilisateur selectionne la base sauvgarde son chemin dans un
    # fichier pour assurer la persistance.
    # C'eest ce fichier qui est lu pour déterminer la base à charger.
    # Autrement qd l'utilisateur change de page via les liens, la base n'est plus chargée.
    if (input$database != "") {
      selected_database(input$database)
      saveRDS(input$database, selected_db_rds)
      showNotification('Database selected.', type='message') # montre la notif que si c'est l'utilisateur qui sélectionne)
    }
    else {
      if (file.exists(selected_db_rds)) {
        selected_db <- readRDS(selected_db_rds)
        selected_database(selected_db)
      }
    }
    
  })
  
  observeEvent(selected_database(), {
    req(selected_database())
    # reinitialise ces valeurs qui ne sont pas forcément communes aux db
    current_locus_tag(NULL)
    current_ortho_group(NULL)
    
    req(file.exists(selected_database())) # si l'utilisateur supprimer le fichier db
    output$current_database <- renderText({selected_database()})
    con(DBI::dbConnect(RSQLite::SQLite(), selected_database()))
  })
  
  
  # Gestion des paramètres de l'URL
  observe({
    query <- getQueryString(session)

    # ?locus_tag=value
    # Change l'onglet courant pour aller sur la bonne page
    # et essaye d'afficher les info pour la valeur donnée
    if ("locus_tag" %in% names(query)) {
      if (!is.null(query$locus_tag)) {
        current_locus_tag(query$locus_tag)
      }
      updateTabItems(session, "sidebarID", selected = "locus_tag")
    }

    if ("strains" %in% names(query)) {
      updateTabItems(session, "sidebarID", selected = "strains")
    }

    if ("ortho_group" %in% names(query)) {
      if (!is.null(query$ortho_group)) {
        #showNotification(query$ortho_group)
        current_ortho_group(query$ortho_group)
      }
      updateTabItems(session, "sidebarID", selected = "ortho_group")
    }

    # Supprime les paramètres de l'URL
    newURL <- paste0(
      session$clientData$url_protocol,
      "//",
      session$clientData$url_hostname,
      ":",
      session$clientData$url_port,
      session$clientData$url_pathname
    )
    updateQueryString(newURL, mode = "replace", session)
  })


  # Validation des inputs
  iv_sigificant_feature <- InputValidator$new()
  iv_sigificant_feature$add_rule("qvalue_threshold", sv_between(0, 1))
  iv_sigificant_feature$add_rule("log2fc_threshold", sv_gte(0))
  iv_sigificant_feature$enable()


  iv_insertions_viz <- InputValidator$new()
  # iv_insertions_viz$add_rule("how_to_handle_rep", sv_in_set(c("sum", "mean")))
  iv_insertions_viz$add_rule("ymax_cutoff", sv_gte(0))
  iv_insertions_viz$add_rule("feature_context", sv_gte(0))
  iv_insertions_viz$enable()

  iv_insertions_viz_ortho <- InputValidator$new()
  iv_insertions_viz_ortho$add_rule("ymax_cutoff", sv_gte(0))
  iv_insertions_viz_ortho$enable()

  




  # PAGE BROWSE BY LOCUS TAG ---------------------------------------------------
  # Cette valeur peut être modifiée en passant un paramètre dans l'url ou en
  # selectionnnant une ligne dans le tableau de features
  current_locus_tag <- reactiveVal(NULL)
  output$current_locus_tag <- renderText({
    current_locus_tag()
  })

  # Cette table permet de lister les features avec leur description
  # Elle permet aussi de sélectionner une feature en particulier
  feature_description_table <- reactive({
    req(con())
    gene_db <- tbl(con(), "gene")
    ortho_group_db <- tbl(con(), "ortho_group")

    gene_db %>%
      left_join(ortho_group_db, by = "locus_tag") %>%
      rename(ortho_group = group) %>%
      collect() # %>%
    # mutate(id = row_number()) # attache un identifiant à chaque ligne
  })

  # Mise en forme
  output$feature_description_table <-
    DT::renderDataTable(
      {
        feature_description_table() %>% mutate(
          ortho_group = if_else(!is.na(ortho_group),
            glue("<a href='?ortho_group={ortho_group}'>{ortho_group}</a>"),
            ortho_group
          )
        )
      },
      escape = FALSE, # pour pouvoir afficher les liens et non les balises.
      server = TRUE,
      options = list(
        pageLength = 5,
        autoWidth = FALSE,
        scrollX = TRUE
      ),
      selection = list(mode = "single", selected = NULL)
    )


  # Quand une feature est sélectionnée on met à jour "current_locus_tag"
  observeEvent(input$feature_description_table_rows_selected, {
    idx <- input$feature_description_table_rows_selected
    lt <- feature_description_table() %>%
      slice(idx) %>%
      pull("locus_tag")
    current_locus_tag(lt)
  })

  # Montre la ligne sélectionné
  # Parce que je n'ai pas trouvé comment sélectionner la bonne ligne
  # quand le locus est passé par l'url
  # https://community.rstudio.com/t/programmatically-update-row-selection-or-other-attributes-of-reactable/47443
  output$selected_locus_tag_table <- DT::renderDataTable({
    req(current_locus_tag())
    lt <- current_locus_tag()
    feature_description_table() %>%
      filter(locus_tag == lt) %>%
       mutate( ortho_group = if_else(!is.na(ortho_group),
                              glue("<a href='?ortho_group={ortho_group}'>{ortho_group}</a>"),
                              ortho_group
                              )
      ) %>%
      datatable(escape = FALSE, options = list(
        dom = "t", # n'affiche que la table
        autoWidth = FALSE,
        scrollX = TRUE
      ))
  })

  # Récupération des prédictions pour la méthode hmm
  hmm_predictions_table <- reactive({
    req(con())
    req(current_locus_tag())
    lt <- current_locus_tag()
    prediction_essentiality_db <-
      tbl(con(), "prediction_essentiality")
    prediction_essentiality_db %>%
      filter(locus_tag == lt) %>%
      select(-c(strain, locus_tag)) %>%
      # select(-c(strain, locus_tag)) %>%
      collect()
  })

  # Mise en forme
  output$hmm_predictions_table <- DT::renderDataTable(
    {
      req(nrow(hmm_predictions_table()) > 0) # il faut au moins une ligne dans le tableau
      datatable(
        hmm_predictions_table(),
        caption = "HMM predictions:",
        selection = "none",
        options = list(
          dom = "t", # n'affiche que la table
          autoWidth = FALSE,
          scrollX = TRUE
        )
      ) %>% formatStyle("call",
        fontWeight = "bold",
        color = "#ffffff",
        backgroundColor = hmm_prediction_color
      )
    },
    server = TRUE
  )

  # Récupération des prédictions pour la méthode resampling
  resampling_predictions_table <- reactive({
    req(con())
    req(current_locus_tag())
    req(iv_sigificant_feature$is_valid())

    lt <- current_locus_tag()
    qvalue_threshold <- input$qvalue_threshold
    log2fc_threshold <- input$log2fc_threshold

    prediction_conditional_essentiality_db <-
      tbl(con(), "prediction_conditional_essentiality")
    prediction_conditional_essentiality_db %>%
      filter(locus_tag == lt) %>%
      select(-c(strain, locus_tag)) %>%
      collect() %>%
      mutate(signif = is_significant(qval, logfc, qvalue_threshold, log2fc_threshold)) # ajout d'une col pour dire si la prediction est significative
  })

  # Mise en forme
  output$resampling_predictions_table <- DT::renderDataTable({
    req(nrow(resampling_predictions_table()) > 0) # il faut au moins une ligne dans le tableau
    table <- resampling_predictions_table() %>%
      add_row(.before = 1, condition = "control") # pour la table soit alignée avec la table des prédictions hmm

    datatable(
      table,
      caption = "Resampling predictions:",
      selection = "none",
      options = list(
        dom = "t",
        autoWidth = FALSE,
        scrollX = TRUE,
        columnDefs = list(list(targets = c(-1), visible = FALSE)) # cacher la dernière colone (signif)
      )
    ) %>% formatStyle("logfc", "signif",
      fontWeight = "bold",
      color = "#ffffff",
      backgroundColor = resampling_prediction_color
    )
  })

  # Pour avoir quelques info supplémentaires sur la feature selectionnée
  # Notamment le début et la fin
  current_feature <- reactive({
    req(con())
    
    req(current_locus_tag())
    lt <- current_locus_tag()

    gene_db <- tbl(con(), "gene")
    feature <- gene_db %>%
      filter(locus_tag == lt) %>%
      collect()

    req(nrow(feature) == 1)

    strain <- feature %>%
      pull("strain") %>%
      head(1)
    start_on_ref <- feature %>%
      pull("start_on_reference") %>%
      head(1)
    end_on_ref <- feature %>%
      pull("end_on_reference") %>%
      head(1)
    list(strain = strain, start_on_ref = start_on_ref, end_on_ref = end_on_ref)
  })

  # table des insertions pour la feature courante.
  # une souche dans une condition peut avoir plusieurs rep.
  # on calcul pour chaque souche et condition la moyenne/somme pour chaque position
  # ex: souche 1, condition A, 2 rep :
  # rep 1 :
  # pos 2, count 5
  # rep2 :
  # pos 2, count7
  # resultat :
  # pos2, mean=6, sum=12
  wiggles_table <- reactive({
    
    req(con())
    req(current_locus_tag())

    strain_crit <- current_feature()$strain
    start_on_ref_crit <- current_feature()$start_on_ref
    end_on_ref_crit <- current_feature()$end_on_ref

    wiggle_db <- tbl(con(), "wiggle")

    wiggle_db %>%
      filter(strain == strain_crit &
        position >= start_on_ref_crit &
        position <= end_on_ref_crit) %>%
      group_by(condition, position) %>%
      summarise(
        mean = mean(count,
          na.rm =
            TRUE
        ),
        sum = sum(count,
          na.rm =
            TRUE
        ),
        .groups = "drop"
      ) %>%
      collect()
  })

  # calcul un certain nb de stat pour la feature courante
  output$insertions_overview_table <- DT::renderDataTable({
    req(wiggles_table())

    start_on_ref_crit <- current_feature()$start_on_ref
    end_on_ref_crit <- current_feature()$end_on_ref

    wiggles_table() %>%
      group_by(condition) %>%
      summarise(
        # TAs = n(),
        saturation = saturation_fun(.data[[input$how_to_handle_rep]]),
        mean = mean(.data[[input$how_to_handle_rep]], na.rm = TRUE),
        sum = sum(.data[[input$how_to_handle_rep]], na.rm = TRUE)
      ) %>%
      datatable(
        caption = "Descriptive statistics:",
        selection = "none",
        options = list(
          dom = "t",
          autoWidth = FALSE,
          scrollX = TRUE
        )
      ) %>%
      formatRound(columns = c("mean", "sum", "saturation"), 2)
  })

  # C'est cette table qui est utilisée pour la visualisation comme elle
  # tient compte du contexte contrairement à wiggles_table
  wiggles_with_context_table <- reactive({
    req(con())
    req(current_locus_tag())

    strain_crit <- current_feature()$strain
    start_on_ref_crit <- current_feature()$start_on_ref
    end_on_ref_crit <- current_feature()$end_on_ref

    wiggle_db <- tbl(con(), "wiggle")

    # pour tenir compte du context
    start_context <- start_on_ref_crit - input$feature_context

    end_context <- end_on_ref_crit + input$feature_context

    wiggle_db %>%
      filter(strain == strain_crit &
        position >= start_context &
        position <= end_context) %>%
      group_by(condition, position) %>%
      summarise(
        mean = mean(count,
          na.rm =
            TRUE
        ),
        sum = sum(count,
          na.rm =
            TRUE
        ),
        .groups = "drop"
      ) %>%
      collect()
  })

  # Prend une dépendance sur deux evenements pour pouvoir mettre à jour
  # ymax cutoff
  ymax_is_to_update <- reactive({
    list(wiggles_with_context_table(), input$how_to_handle_rep)
  })

  # Met à jour ymax cutoff en prenant la valeur maximale (sum|mean) + une marge
  observeEvent(ymax_is_to_update(), {
    req(wiggles_with_context_table())
    table <- wiggles_with_context_table()
    ymax <- as.integer(max(table[[input$how_to_handle_rep]]))
    if (is.na(ymax)) { # gere les cas ou la feature n'a pas de site TA
      ymax <- 0
    }
    updateNumericInput(session, "ymax_cutoff", value = ymax + 20)
  })

  # visualise les insertions pour la feature courante et toutes les conditions
  output$insertions_plot <- renderPlotly({
    req(wiggles_with_context_table())
    req(iv_insertions_viz$is_valid())

    table <- wiggles_with_context_table()
    strain_crit <- current_feature()$strain
    start_on_ref_crit <- current_feature()$start_on_ref
    end_on_ref_crit <- current_feature()$end_on_ref

    ymax <- input$ymax_cutoff
    plot_feature_insertions(table, ymax, start_on_ref_crit, end_on_ref_crit, input$how_to_handle_rep) %>% layout(hovermode = "x unified")
  })



  # PAGE DISCOVER LOCUS TAG ----------------------------------------------------
  # cette table est là pour découvrir des features intéressantes
  # à partir des prédictions faites par HMM
  hmm_overall_prediction_table <- reactive({
   req(con())
   prediction_essentiality_db <- tbl(con(), "prediction_essentiality")
   ortho_group_db <- tbl(con(), "ortho_group")
   data <- prediction_essentiality_db %>%
       select(strain, condition, locus_tag, call) %>%
       left_join(ortho_group_db, by = "locus_tag") %>%
       rename(ortho_group = group) %>%
       collect()

   data <- data %>% pivot_wider(names_from = condition, values_from = call) %>%
       group_by(ortho_group) %>%
       mutate(locus_tag_number_in_ortho_group = length(na.omit(ortho_group)), .after = ortho_group) %>%
       ungroup() %>%
       arrange(desc(locus_tag_number_in_ortho_group), ortho_group)
   })

  # construit la liste des conditions disponibles pour les prédictions hmm
  hmm_conditions <- reactive({
    req(con())
    prediction_essentiality_db <- tbl(con(), "prediction_essentiality")
    prediction_essentiality_db %>%
      select(condition) %>%
      distinct() %>%
      collect() %>%
      pull(condition)
  })

  # filter et mise en forme de la table des predictions hmm
  output$hmm_overall_prediction_table <- DT::renderDataTable({
    req(hmm_overall_prediction_table())
    req(hmm_conditions())

    hmm_conditions_val <- hmm_conditions()
    data <- hmm_overall_prediction_table() %>% mutate(
      locus_tag = glue("<a href='?locus_tag={locus_tag}'>{locus_tag}</a>"),
      ortho_group = if_else(!is.na(ortho_group),
                            glue("<a href='?ortho_group={ortho_group}'>{ortho_group}</a>"),
                            ortho_group
      )
    )
    if (input$ignore_boring_rows) {
      data <- data %>% filter(if_any(hmm_conditions_val, ~ .x != "NE"))
    }
    datatable(data,
      escape = FALSE,
      #extensions = "RowGroup",
      options = list(
        #rowGroup = list(dataSrc = c(3)), # groupe les locus_tag par ortho_group
        # ATTENTION : les groupes se font sur la page courante
        # Si un locus_tag est dans une page suivante, il ne sera pas mis dans le groupe
        #columnDefs = list(list(visible = TRUE, targets = c(3))), # pour que la col ortho_group soit visible meme après le regroupement
        dom = "ltip",
        # searchHighlight = TRUE,
        search = list(regex = TRUE),
        autoWidth = FALSE,
        scrollX = TRUE
      ),
      selection = "none",
      filter = "top"
    ) %>%
      formatStyle(
        hmm_conditions_val,
        fontWeight = "bold",
        color = "#ffffff",
        backgroundColor = hmm_prediction_color
      )
  })



  # cette table est là pour découvrir des features intéressantes
  # à partir des prédictions faites par RESAMPLING
  resampling_overall_prediction_table <- reactive({
    req(con())
    prediction_conditional_essentiality_db <- tbl(con(), "prediction_conditional_essentiality")
    ortho_group_db <- tbl(con(), "ortho_group")

    data <- prediction_conditional_essentiality_db %>%
      select(strain, condition, locus_tag, qval, logfc) %>%
      left_join(ortho_group_db, by = "locus_tag") %>%
      rename(ortho_group = group) %>%
      collect()
  })

  # construit la liste des conditions disponibles pour les prédictions resampling
  resampling_conditions <- reactive({
    req(con())
    prediction_conditional_essentiality_db <- tbl(con(), "prediction_conditional_essentiality")
    prediction_conditional_essentiality_db %>%
      select(condition) %>%
      distinct() %>%
      collect() %>%
      pull(condition)
  })

  output$resampling_overall_prediction_table <- DT::renderDataTable({
    req(resampling_overall_prediction_table())
    req(resampling_conditions())
    req(iv_sigificant_feature$is_valid())


    # showNotification("format")
    resampling_conditions_val <- resampling_conditions()

    qvalue_threshold <- input$qvalue_threshold
    log2fc_threshold <- input$log2fc_threshold


    data <- resampling_overall_prediction_table() %>%
      mutate(signif = is_significant(qval, logfc, qvalue_threshold, log2fc_threshold)) %>% # ajout d'une col pour dire si la prediction est significative
      # filter(signif != "unsignificant") %>%
      select(strain, locus_tag, ortho_group, condition, signif) %>%
      pivot_wider(names_from = condition, values_from = signif) %>%
      group_by(ortho_group) %>%
      mutate(locus_tag_number_in_ortho_group = length(na.omit(ortho_group)), .after = ortho_group) %>%
      ungroup() %>%
      arrange(desc(locus_tag_number_in_ortho_group)) %>%
      mutate(locus_tag = glue("<a href='?locus_tag={locus_tag}' >{locus_tag}</a>"))


    if (input$ignore_boring_rows) {
      data <- data %>% filter(if_any(resampling_conditions_val, ~ .x != "unsignificant"))
    }

    datatable(data,
      escape = FALSE,
       #extensions = "RowGroup",
      options = list(
        #rowGroup = list(dataSrc = c(3)),
        #columnDefs = list(list(visible = TRUE, targets = c(3))),
        dom = "ltip",
        search = list(regex = TRUE),
        autoWidth = FALSE,
        scrollX = TRUE
        # lengthMenu = c(10,50, 100, 200, 500),
        # pageLength = 10
      ),
      selection = "none",
      filter = "top"
    ) %>%
      formatStyle(
        resampling_conditions_val,
        fontWeight = "bold",
        color = "#ffffff",
        backgroundColor = resampling_prediction_color
      )
  })


  # PAGE STRAINS ---------------------------------------------------------------
  # recupere toutes les features pour chaque souche et calcul le #features
  strain_and_feature_table <- reactive({
    req(con())
    strain_db <- tbl(con(), "strain")
    gene_db <- tbl(con(), "gene")
    gene_db %>%
      mutate(length = end_on_reference - start_on_reference + 1)

    strain_db %>%
      select(-sequence) %>%
      left_join(gene_db, by = "strain") %>%
      group_by(strain) %>%
      mutate(feature_number = n()) %>%
      ungroup() %>%
      rename(feature_ta_number = ta_number.y, ta_number = ta_number.x) %>%
      collect()
  })

  strain_overview_table <- reactive({
    req(strain_and_feature_table())
    strain_and_feature_table() %>%
      select(strain, genus, length, gc_percent, ta_number, feature_number) %>%
      distinct()
  })

  # mise en forme
  output$strain_overview_table <- DT::renderDataTable({
    req(strain_overview_table())
    strain_overview_table() %>%
      datatable(options = list(
        dom = "ltp",
        pageLength = 5,
        autoWidth = FALSE,
        scrollX = TRUE
      ), selection = list(mode = "single", selected = NULL)) %>%
      formatPercentage("gc_percent", 2) %>%
      formatRound(c("length", "ta_number", "feature_number"), 0)
  })

  # Qd une souche est sélectionnée on affiche la distribution du nb de sites TA
  # de toutes les features
  observeEvent(input$strain_overview_table_rows_selected, {
    idx <- input$strain_overview_table_rows_selected
    # showNotification(idx)
    strain_crit <- strain_overview_table() %>%
      slice(idx) %>%
      pull(strain)

    output$view <- renderPlotly({
      table <- strain_and_feature_table() %>%
        filter(strain == strain_crit) %>%
        group_by(feature_ta_number) %>%
        summarize(frequency = n())

      table %>% plot_ly(x = ~feature_ta_number, y= ~frequency, type = "bar", marker = list(color = "#848FA2")) %>%
        layout(title = "Distribution of the number of TA sites by feature",
                     xaxis = list(title = "TA number by feature"),
                     yaxis = list(title = "Frequency")
                     )
    })
  })
  
  output$orthologous_groups_number <- renderText({
    req(con())
     ortho_group_db <- tbl(con(), "ortho_group")
     ortho_group_db %>% select(group) %>% distinct() %>% collect() %>% nrow()
   })
  output$orthologous_groups_distribution <- renderPlotly({
    req(con())
    ortho_group_db <- tbl(con(), "ortho_group")
    table <- ortho_group_db %>% group_by(group) %>%  summarize(feature_number = n()) %>% ungroup() %>% collect()

    table %>% plot_ly(x = ~ feature_number, type="histogram", marker = list(color = "#848FA2")) %>%
      layout(title = "Distribution of the number of features by orthologous group",
             xaxis = list(title = "Feature number by group"),
             yaxis = list(title = "Frequency")
      )
  })
  # PAGE ORTHO_GROUP------------------------------------------------------------
  # Quick & dirty
  # TODO ne pas faire la requete en base à chaque fois
  # Il faut séparer la récupération des données et les filtres
  current_ortho_group <- reactiveVal(NULL)
  output$ortho_group_insertions_plot <- renderPlotly({
    req(con())
    req(current_ortho_group())

    gene_db <- tbl(con(), "gene")
    ortho_group_db <- tbl(con(), "ortho_group")
    ortho_group_crit <- current_ortho_group()
    feature_table <- gene_db %>%
      left_join(ortho_group_db, by = "locus_tag") %>%
      rename(ortho_group = group) %>%
      filter(ortho_group == ortho_group_crit) %>%
      collect()

    req(nrow(feature_table) >=1)

    figures <- list()
    for(i in 1:nrow(feature_table)) {
      strain_crit <- feature_table %>% slice(i) %>% pull(strain) #%>% first()
      start_on_ref_crit <- feature_table %>% slice(i) %>% pull(start_on_reference)
      end_on_ref_crit <- feature_table %>% slice(i) %>% pull(end_on_reference)

      wiggle_db <- tbl(con(), "wiggle")
      wiggle_table <- wiggle_db %>% filter(strain == strain_crit &
                                             position >= start_on_ref_crit &
                                             position <= end_on_ref_crit) %>% collect()
      wiggle_table <- wiggle_table %>% group_by(condition, position) %>%
        summarise(
          mean = mean(count,
                      na.rm =
                        TRUE
          ),
          sum = sum(count,
                    na.rm =
                      TRUE
          ),
          .groups = "drop"
        ) %>%
        collect()

      ymax <- input$ymax_cutoff_ortho


      # Its a trick to avoid to duplicate the legend for each iteration
      if (i == 1){
        show_legend = T
      }else{
        show_legend = F
      }

      fig <- plot_feature_insertions(wiggle_table, ymax, start_on_ref_crit, end_on_ref_crit, input$how_to_handle_rep, show_legend=show_legend, make_it_relative = T, strain=strain_crit)

      figures[[ strain_crit ]] <- fig
    }

  # combine all the plots in one figure
  overvall_fig <- subplot(figures, nrows = length(figures), shareX = TRUE, titleY=TRUE, margin = .005) %>%
    layout(title = list(text = ""),
           hovermode = "x unified")#,
           #plot_bgcolor='#e5ecf6')
  })
}
# -----------------------------------------------------------------------------#
# APP
# -----------------------------------------------------------------------------#

shinyApp(ui, server, enableBookmarking = "disable")
