library(shiny)
library(shinydashboard)
library(shinyFiles)
library(tidyverse)
library(plotly)
# library(reactlog)
library(DT)
library(shinybusy)
# library(shinyWidgets)
library(glue)
library(formattable) # color_bar
library(shinyjs) # click
# library(profvis) #https://rstudio.github.io/profvis/examples.html#example-3---profiling-a-shiny-application
# profvis({
#  runApp('pipeline/qc_shiny')
# })

PLAYGROUND <- file.path(fs::path_home(), "playground")
addResourcePath("files", PLAYGROUND)

# Modules
source("modules/input_module.R", local = TRUE)
source("modules/run_module.R", local = TRUE)
source("modules/process_reads_module.R", local = TRUE)
source("modules/core_analysis_module.R", local = TRUE)
source("modules/condi_analysis_module.R", local = TRUE)
source("modules/normalization_module.R", local = TRUE)
source("modules/create_database_module.R", local = TRUE)
source("modules/afterwards_module.R", local = TRUE)

# Load some useful functions
path_to_functions_script = Sys.which("functions.R")
source(path_to_functions_script)


# UI
ui <- dashboardPage(
  dashboardHeader(title = "TnSeek"),
  dashboardSidebar(
    sidebarMenu(id="navbar",
      menuItem(
        "Home",
        tabName = "home",
        icon = icon("home", lib = "glyphicon")
      ),
      menuItem("1 - Reads processing", tabName = "process_reads"),
      menuItem("2 - Normalization", tabName = "normalize"),
      menuItem("3 - Core genome", tabName = "core_genome"),
      menuItem("4 - Conditionally essential genes", tabName = "condi_analysis"),
      menuItem("5 - Create database", tabName = "create_database")
    ) # ,
    # hr(),
    # textInput("project", "Current project name", placeholder = "meaningful name")
  ),
  dashboardBody(
    useShinyjs(),
    tags$head(tags$style('[id$="-progress"]{overflow-y:scroll; max-height: 500px; font-family: monospace,monospace; background-color: #272822; color: #f8f8f2;}')),
    tags$head(tags$style('[id$="-command"]{overflow-y:scroll; max-height: 500px; font-family: monospace,monospace; background-color: #272822; color: #f8f8f2;}')),
    add_busy_bar(color = "red", height = "8px"),
    tabItems(
      tabItem(
        tabName = "home",
        h2("Welcome")
      ),
      tabItem(
        tabName = "process_reads",
        processReadsUI("process_reads"),
        afterwardsReportUI("process_reads"),
        afterwardsDownstreamUI("process_reads")
      ),
      tabItem(
        tabName = "normalize",
        normalizationUI("normalize"),
        afterwardsReportUI("normalize"),
        afterwardsDownstreamUI("normalize")
      ),
      tabItem(
        tabName = "core_genome",
        coreAnalysisUI("core_genome"),
        afterwardsReportUI("core_genome"),
        afterwardsDownstreamUI("core_genome")
      ),
      tabItem(
        tabName = "condi_analysis",
        condiAnalysisUI("condi_analysis"),
        afterwardsReportUI("condi_analysis"),
        afterwardsDownstreamUI("condi_analysis")
      )
      ,
      tabItem(
        tabName = "create_database",
        createDatabaseUI("create_database")
      )
    )
  )
)

# Server
server <- function(input, output, session) {
  # check that the playground directory exists on the server
  if (!file.exists(PLAYGROUND)) {
    stop(paste("The PLAYGROUND directory does not exists on this server:", PLAYGROUND))
  }

  process_reads_data <- processReadsServer("process_reads")
  normalize_data <- normalizationServer("normalize")
  core_data <- coreAnalysisServer("core_genome")
  condi_data <- condiAnalysisServer("condi_analysis")
  createDatabaseServer("create_database")
 
  # Pour être en mesure d'afficher les liens vers les prochaines analyses,
  # le module afterwards doit avoir accès aux fichiers d'entré de chaque module
  data_list = reactive({
    list(
    process_reads = process_reads_data,
    normalize = normalize_data,
    core_genome = core_data,
    condi_analysis = condi_data
  )})
  
  # Pour être en mesure de changer d'onglet, definition d'une reactiveVal
  # qui sera passée au module afterwards
  selectedTab <- reactiveVal(NULL)
  observeEvent(selectedTab(), {
    if (selectedTab() %in% c("process_reads", "normalize", "core_genome", "condi_analysis"))
      updateTabItems(session, "navbar", selected = selectedTab())
  })
  
  afterwardsServer("process_reads", process_reads_data, data_list, selectedTab)
  afterwardsServer("normalize", normalize_data, data_list, selectedTab)
  afterwardsServer("core_genome", core_data, data_list, selectedTab)
  afterwardsServer("condi_analysis", condi_data, data_list, selectedTab)
}

shinyApp(ui, server)
