# This module is not intended to be reusable.
# This module is intended to avoid namespace collision between parameters in different part of the app (process-read, core..)

# UI
condiAnalysisUI <- function(id) {
  ns <- NS(id)
  fluidRow(
    box(
      width = 12,
      title = "Inputs",
      status = "primary",
      tags$p("Enter your input file (normalized count & annotations, TSV format, required):"),
      inputFileUI(ns("inputs-table")),
      br(),
      tags$p("Enter your input file (orthology table, TSV format, optional):"),
      inputFileUI(ns("ortho-table"))
    ),
    uiOutput(ns("condi_analysis_ui"))
  )
}

# Server
condiAnalysisServer <- function(id) {
  moduleServer(id, function(input, output, session) {
    inputs_table_path <- inputFileServer("inputs-table")
    ortho_table_path <- inputFileServer("ortho-table")

    project_name <- reactive({
      req(inputs_table_path())
      fs::path_split(fs::path_rel(inputs_table_path(), start = PLAYGROUND))[[1]][1]
    })

    outdir <- reactive({
      outdir <- file.path(PLAYGROUND, project_name(), "condi_analysis")
    })


    command <- reactive({
      inputs_table_path_val <- inputs_table_path()
      ortho_table_path_val <- ortho_table_path()
      samples_number <- input$samples_number

      ortho_grps_args <- ''
      if (! is.null(ortho_table_path_val)) {
        ortho_grps_args <- paste0("--ortho-grps ",  '"', ortho_table_path_val,  '" \n')
      }

      pseudocount <- input$pseudocount
      ignore_n <- input$ignore_n
      ignore_c <- input$ignore_c

      exclude_zeros <- ""
      if (!input$include_zeros) {
        exclude_zeros <- " --exclude-rows-with-zero"
      }

      adaptative <- ""
      if (input$adaptative_resampling) {
        adaptative <- " --adaptative-resampling"
      }

      outdir_val <- outdir()
      
      verbose_option <- ''
      if (input$verbose) {
        verbose_option <- ' --verbose'
      }

      glue(
        "tnphyto.py condi-analysis -o \"{outdir_val}\" \n",
        "-i \"{inputs_table_path_val}\" \n",
        "{ortho_grps_args}",
        "--samples-number {samples_number} \n",
        "--pseudocount {pseudocount} \n",
        "--ignoreN {ignore_n} \n",
        "--ignoreC {ignore_c} \n",
        "{exclude_zeros}{adaptative}{verbose_option} \n"
      )
    })

    output$command <- renderText(
      command()
    )
    runBackgroundJobServer("condi-analysis", command, outdir)

    output$condi_analysis_ui <- renderUI({
      # This UI will be rendered when the inputs are set by the user
      # Basically, it contains the parameters used for the analysis and a button to run it.
      req(inputs_table_path())
      tagList(
        box(
          width = 12,
          title = "Parameters",
          status = "primary",
          collapsible = T,
          fluidRow(
            column(
              width = 5,
              numericInput(session$ns("samples_number"), label = "Number of samples", value = 10000),
              helpText(
                "The number of samples (permutations) to perform."
              ),
              checkboxInput(session$ns("adaptative_resampling"), label = "Adaptative resampling (faster).", value = FALSE),
              helpText(
                "accelerates the calculation by terminating early for genes which are likely not significant. This dramatically speeds up the computation at the cost of less accurate estimates for those genes that terminate early (i.e. deemed not significant). "
              )
            ),
            column(
              width = 5,
              offset = 1,
              numericInput(session$ns("pseudocount"), label = "Pseudocount", value = 0),
              helpText(
                "Used in calculating log-fold-change. Useful to dampen the effects of small counts which may lead to deceptively high LFC."
              ),
              checkboxInput(session$ns("include_zeros"), label = "Include sites with all zeros", value = TRUE),
              helpText("Select to include sites that are zero. This is the preferred behavior, however, unselecting this (thus ignoring sites that) are zero accross all dataset (i.e. completely empty), is useful for decreasing running time."),
            )
          ),
          hr(),
          fluidRow(
            column(
              width = 5,
              sliderInput(
                session$ns("ignore_n"),
                label = "Ignore N-terminus %",
                min = 0,
                max = 20,
                step = 1,
                value = 0
              ),
              helpText("Ignores a fraction of the ORF, beginning at the N-terminal end. Useful for ignoring read-counts that may occur at the terminal ends, even though they do not truly disrupt a genes function."),
            ),
            column(
              width = 5,
              offset = 1,
              sliderInput(
                session$ns("ignore_c"),
                label = "Ignore c-terminus %",
                min = 0,
                max = 20,
                step = 1,
                value = 0
              ),
              helpText("Ignores a fraction of the ORF, beginning at the C-terminal end. Useful for ignoring read-counts that may occur at the terminal ends, even though they do not truly disrupt a genes function."),
            )
          ),
          fluidRow(
            column(
              width = 5,
              tags$label("Verbosity:"),
              checkboxInput(
                session$ns("verbose"),
                label="verbose"
              ),
              helpText("Display more informations.")
            )
          )
        ),
        box(
          width = 12,
          title = "Command",
          status = "primary",
          collapsible = T,
          collasped = T,
          verbatimTextOutput(session$ns("command")),
          runButtonUI(session$ns("condi-analysis"), "Compute conditionnal genes"),
        ),
        box(
          width = 12,
          title = "Logs",
          status = "primary",
          collapsible = T,
          collapsed = T,
          runLogUI(session$ns("condi-analysis"))
        )
      )
    })
    list(
      inputs_table = inputs_table_path, # to be able to set this value from outside
      outdir = outdir,
      next_steps = c()
    )
  })
}
