# This module is not intended to be reusable.
# This module is intended to avoid namespace collision between parameters in different part of the app (process-read, core..)

# UI
processReadsUI <- function(id) {
  ns <- NS(id)
  fluidRow(
    box(
      width = 12,
      title = "Inputs",
      status = "primary",
      tags$p("Enter your input file (TSV file with reads & annotations):"),
      inputFileUI(ns("inputs_table"))
    ),
    uiOutput(ns("process_reads_ui"))
  )
}

# Server
processReadsServer <- function(id) {
  moduleServer(id, function(input, output, session) {
    inputs_table_path <- inputFileServer("inputs_table")

    project_name <- reactive({
      req(inputs_table_path())
      fs::path_split(fs::path_rel(inputs_table_path(), start = PLAYGROUND))[[1]][1]
    })

    outdir <- reactive({
      outdir <- file.path(PLAYGROUND, project_name(), "process_reads")
    })

    command <- reactive({
      # validate(
      #  need(grepl("^[ACGT]*$", input$transposon), "That's not a valid transposon sequence")
      # )
      transposon <- input$transposon
      cut <- input$cut
      read_length <- input$read_length
      read_min_length <- read_length[1]
      read_max_length <- read_length[2]

      min_overlap <- input$min_overlap
      max_error_rate <- input$max_error_rate
      inputs_table_path_val <- inputs_table_path()
      outdir_val <- outdir()
      
      verbose_option <- ''
      if (input$verbose) {
        verbose_option <- '--verbose'
      }

      glue(
        "tnphyto.py process-reads -i \"{inputs_table_path_val}\" -o \"{outdir_val}\" \n",
        "--transposon \"{transposon}\" \n",
        "--cut {cut} \n",
        "--max-error-rate {max_error_rate} \n",
        "--min-overlap {min_overlap} \n",
        "--minimum-length {read_min_length} \n",
        "--maximum-length {read_max_length} \n",
        "{verbose_option} \n"
      )
    })

    output$command <- renderText(
      command()
    )

    runBackgroundJobServer("runner", command, outdir)
    
    output$process_reads_ui <- renderUI({
      # This UI will be rendered when the inputs are set by the user
      # Basically, it contains the parameters used for the analysis and a button to run it.
      if (is.null(inputs_table_path())) {
        return(c())
      }

      tagList(
        box(
          width = 12,
          title = "Parameters",
          status = "primary",
          collapsible = T,
          fluidRow(
            column(
              width = 5,
              textInput(session$ns("transposon"),
                label = "Transposon:", value =
                  "ACAGGYTGGATGATAAGTCCCCGGTCT"
              ),
              helpText("Sequence of the transposon to look for at the 3' end of the reads"),
              sliderInput(
                session$ns("read_length"),
                label = "Read length:",
                min = 10,
                max = 100,
                step = 1,
                value = c(15, 20)
              ),
              helpText(
                "Discard processed reads that are shorter/greater than min/max LENGTH"
              ),
            ),
            column(
              width = 5,
              offset = 1,
              numericInput(
                session$ns("min_overlap"),
                label = "Min overlap:",
                min = 3,
                value = 7
              ),
              helpText("The required MINLENGTH overlap between read and transposon"),
              sliderInput(
                session$ns("max_error_rate"),
                label = "Max error rate:",
                min = 0,
                max = 0.99,
                step = 0.01,
                value = 0.05
              ),
              helpText("The level of error used to search the transposon")
            ),
            column(
              width = 5,
              numericInput(
                session$ns("cut"),
                label = "Cut:",
                min = 0,
                value = 0
              ),
              helpText("Number of bases to remove from the beginning of each read.")
            ),
            column(
              width = 5,
              offset = 1,
              tags$label("Verbosity:"),
              checkboxInput(
                session$ns("verbose"),
                label="verbose"
              ),
              helpText("Display more informations.")
            ),
          ),
        ),
        box(
          width = 12,
          title = "Command",
          status = "primary",
          collapsible = T,
          collapsed = F,
          verbatimTextOutput(session$ns("command")),
          runButtonUI(session$ns("runner"), "Process reads"),
        ),
        box(
          width = 12,
          title = "Logs",
          status = "primary",
          collapsible = T,
          collapsed = T,
          runLogUI(session$ns("runner"))
        )
      )
    })
     list(
       inputs_table = inputs_table_path, # to be able to set this value from outside
       outdir = outdir,
       next_steps = c('normalize')
     )
  })
  
}
