

# UI
runButtonUI <- function(id, label = 'Run') {
  ns <- NS(id)
  actionButton(ns("run"), label)
}

runLogUI <- function(id) {
  ns <- NS(id)
  verbatimTextOutput(ns("progress"))
}


# Server
runBackgroundJobServer <- function(id, command, outdir, not_ok_if_outdir_exists=T) {
  moduleServer(id, function(input, output, session) {
    ## file to write progress
    tfile <- tempfile(fileext = ".log")
    ## reactive values
    r <- reactiveValues(progress_mtime = -1)
    
    observeEvent(input$ok, {
      unlink(outdir(), recursive = TRUE)
      showNotification("Files deleted")
      removeModal()
    })
    observeEvent(input$cancel, {
      removeModal()
    })

    observeEvent(input$run, {
      if (is.null(r$bg_process)) {
        if ((not_ok_if_outdir_exists) && (file.exists(outdir()))) {
          modal_confirm <- modalDialog(
            "In order to process, you have to DELETE the output directory first:",
            fs::path_rel(outdir(), start=PLAYGROUND),  
            title = "Output directory already exists",
            footer = tagList(
              actionButton(session$ns("cancel"), "Cancel"),
              actionButton(session$ns("ok"), "Delete", class = "btn btn-danger")
            )
          )
          showModal(modal_confirm)
        }
        else {
          command_txt <- gsub("\n", "", command())
          wrapped_command <-
            glue(
              "/usr/bin/bash -c 'eval \"$(conda shell.bash hook)\" && conda activate tnphyto && {command_txt}'"
            )
          
          r$bg_process <- callr::r_bg(
            func = function(command) {
              system(command)
            },
            args = list(command = wrapped_command),
            stdout = tfile,
            stderr = "2>&1",
            supervise = TRUE
          )
          ## start polling bg process
          r$poll <- TRUE
          pid <- r$bg_process$get_pid()
          showNotification(glue("Running. PID{pid}."),
                           duration = NULL,
                           type = "message")
        }
      }
      else {
        pid <- r$bg_process$get_pid()
        #status <- r$bg_process$get_status()
        showNotification(
          glue(
            "Please, wait until the end of the current job first. PID={pid}."
          ),
          duration = 10,
          type = "warning"
        )
      }
    })
    ## observe status of bg process
    observe({
      req(r$bg_process, r$poll)
      pid <- r$bg_process$get_pid()
      print(r$bg_process$is_alive())
      print(r$bg_process$poll_io(0)["process"])
      
      ## keep re-executing observer as
      ## long as bg process is running
      invalidateLater(millis = 1000, session)
      ## read current progress if file is modified
      mtime <- file.info(tfile)$mtime
      if (mtime > r$progress_mtime) {
        r$progress <- readLines(tfile)
        r$progress_mtime <- mtime
      }
      
      if (!r$bg_process$is_alive()) {
        if (r$bg_process$get_result() != 0) {
          showNotification(
            glue(
              "An error occured. See the logs for more details. PID={pid}."
            ),
            duration = NULL,
            type = "error"
          )
        }
        else {
          showNotification(glue("Success! PID={pid}."),
                           duration = NULL,
                           type = "message")
        }
        print(paste("status", r$bg_process$get_exit_status()))
        r$poll <- FALSE ## stop polling bg process
        r$bg_process <- NULL

        # copy log file to outputdir
        if (file.exists(tfile) && dir.exists(outdir())) {
          file.copy(tfile, outdir())
        }
      }
    })
    ## print progress
    output$progress <- renderText({
      req(r$progress)
      paste(r$progress, collapse = "\n")
    })
  })
}
